#!/usr/local/bin/python
#----------------------------------------------------------------------
# Copyright (c) 2010-2011 USC/ISI
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and/or hardware specification (the "Work") to deal in the
# Work without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Work, and to permit persons to whom the Work is furnished to do so,
# subject to the following conditions: The above copyright notice and this
# permission notice shall be included in all copies or substantial portions of
# the Work.
# 
# 
# THE WORK IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE WORK OR THE USE OR OTHER DEALINGS
# IN THE WORK.
#----------------------------------------------------------------------
from geni.am import ReferenceAggregateManager as BaseReferenceAggregateManager
from abac.cred_util import CredentialVerifier, create_delegation, \
	GID_string_to_ABAC_ID_string, clean_attr

class ReferenceAggregateManager(BaseReferenceAggregateManager):
    def __init__(self, root_dir, cred_dir):
	'''
	Initialize the base GENI AM and then overwrite the CredentialVerifier
	with an ABAC CredentialVerifier.  (Also initialize the know_urns list
	for the poor-man's RT1 hack.
	'''
	BaseReferenceAggregateManager.__init__(self, root_dir)
	self._cred_verifier = CredentialVerifier(cred_dir)
	self.known_urns = []
	self.need_ListResources_priv = True

    def append_proofs(self, rv):
	'''
	If there is room for an ABAC proof in the response, append one,
	identified as geni_proof.
	'''
	principal = GID_string_to_ABAC_ID_string(self._server.pem_cert)
	proof = self._cred_verifier.get_proof(principal)
	if proof and isinstance(rv, dict):
	    rv['geni_proof'] = [ p.to_xml() for p in proof ]
	return rv

    def learn_urn(self, urn):
	'''
	Poor man's RT1.  Add a delegation credential to the ABAC policy
	indicating that we respect a clearinghouses credential for the
	operations on this slice URN as our own.  Note that this does not grant
	any permissions, it just simulates the RT1 semantics of
	createsliver(URN) that have been delegated to clearinghouses.  More
	details in the write-up.
	'''
	if urn not in self.known_urns:
	    c_urn = clean_attr(urn)
	    creds = []
	    for p in ('renewsliver', 'createsliver', 'deleteslice', 
		    'getsliceresources', 'shutdown'):
		creds.append(create_delegation(
		    trust_delegation='clearinghouse.%s_%s' % (p, c_urn),
		    target_attr = '%s_%s' % (p, c_urn),
		    issuer_keyfile=self.key, issuer_certfile=self.certfile))
	    self._cred_verifier.import_creds_binary(creds)
	    self.known_urns.append(urn)

    def set_server(self, s):
	'''
	This allows the ABAC services to operate in the GENI-AM framework
	without changes to the GENI-AM code.  It essentially passes some
	parameters between the ABAC server class and the ABAC AM class that are
	accessible differently from the server classes that the GENI-AM uses.
	Additionally the issuer and key used to enable learn_urn are stored.
	This is called by the abac.server when it it initialized.
	'''
	self._server = s
	self._cred_verifier.set_me(pem=s.me_pem)
	self.certfile = s.me_certfile
	self.key = s.me_key

    # Theses call the GENI AM and append the ABAC certificate information if
    # possible.
    def ListResources(self, credentials, options):
	rv = BaseReferenceAggregateManager.ListResources(self, credentials, 
	    options)
	return self.append_proofs(rv)

    def CreateSliver(self, slice_urn, credentials, rspec, users):
	self.learn_urn(slice_urn)
	rv = BaseReferenceAggregateManager.CreateSliver(self, slice_urn, 
	    credentials, rspec, users)
	return self.append_proofs(rv)

    def DeleteSliver(self, slice_urn, credentials):
	self.learn_urn(slice_urn)
	rv = BaseReferenceAggregateManager.DeleteSliver(self, slice_urn, 
	    credentials)
	return self.append_proofs(rv)

    def SliverStatus(self, slice_urn, credentials):
	self.learn_urn(slice_urn)
	rv = BaseReferenceAggregateManager.SliverStatus(self, slice_urn, 
	    credentials)
	return self.append_proofs(rv)


    def RenewSliver(self, slice_urn, credentials, expiration_time):
	self.learn_urn(slice_urn)
	rv = BaseReferenceAggregateManager.RenewSliver(self, slice_urn, 
	    credentials, expiration_time)
	return self.append_proofs(rv)

    def Shutdown(self, slice_urn, credentials):
	self.learn_urn(slice_urn)
	rv = BaseReferenceAggregateManager.Shutdown(self, slice_urn, 
	    credentials)
	return self.append_proofs(rv)



