#----------------------------------------------------------------------
# Copyright (c) 2008-2010 USC/ISI
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and/or hardware specification (the "Work") to deal in the
# Work without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Work, and to permit persons to whom the Work is furnished to do so,
# subject to the following conditions: The above copyright notice and this
# permission notice shall be included in all copies or substantial portions of
# the Work.
# 
# 
# THE WORK IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE WORK OR THE USE OR OTHER DEALINGS
# IN THE WORK.
#----------------------------------------------------------------------
'''
Credential creation and verification utilities.
'''

# This has the same basic structure to allow the CredentialValidator and
# make_credential routines to plug into the existing code, but they support
# ABAC which makes for some differences.

import os
import os.path
import logging
import xmlrpclib
import sys
import datetime
import subprocess
import re

import ABAC
from proof import proof
from tempfile import mkdtemp

from string import join, printable

from pyasn1.codec.der import decoder
from M2Crypto import X509
from M2Crypto.EVP import MessageDigest

import sfa.trust.credential as cred
import sfa.trust.gid as gid
import sfa.trust.rights as rights

def _find_exec(n, p):
    '''
    Find an executable (n) in path tuple (p).  The first executable is
    returned, or None if none is found.  Used at module initialization time to
    find creddy and openssl.
    '''
    for d in p:
	path = "%s/%s" % (d, n)
	if os.access(path, os.X_OK):
	    return path
    else:
	return None


creddy_path = _find_exec('creddy', ('/usr/bin', '/usr/local/bin'))
openssl_path = _find_exec('openssl', ('/usr/bin', '/usr/local/bin'))

class CredentialVerifier(object):
    """
    Credential Verifier to validate ABAC credentials.
    """
    # The cred_dir is a directory of ABAC certificates that encode the
    # manager's policy.
    def __init__(self, cred_dir):
	'''
	Create the Validator and initilaize it with credentials from cred_dir.
	Also a dictionary of proofs is created so that the ABAC AM can find the
	last proof (or partial proof) generated by a call on behalf of this
	principal.
	'''
        self.logger = logging.getLogger('cred-verifier')
	self.context = ABAC.Context()
	self.proofs = { }
	if cred_dir:
	    if os.path.isdir(cred_dir):
		self.load_creds(self.context.load_id_file, 
			self.context.load_attribute_file,
			["%s/%s" % (cred_dir, fn) \
				for fn in os.listdir(cred_dir)],
			self.logger)
	    else:
		raise Exception("%s is not a directory" % cred_dir)
	else:
	    raise Exception("No certificate directory provided!?")

    def set_me(self, pem):
	'''
	Set the CredentialVerifier's ABAC principal to the ABAC ID derived from
	the given X.509 identity certificate.
	'''
	self.me = GID_string_to_ABAC_ID_string(pem)

    @staticmethod
    def load_creds(load_id, load_attribute, creds, log=None):
	'''
	Helper function to load a list of ABAC credentials into a given
	context.  Either the DER or PEM encoded certificates or the files
	containing them can be passed, assuming that the load_id and
	load_attribute functions are bound to the proper import functions
	(e.g., load_id_file for files for load_id_chunk for binaries).  These
	functions muct be bound to the context that is importing the
	credentials.  
	'''
	def show_cred(c):
	    '''
	    Printing binaries to the log is not very helpful.  Any unprintable
	    credential is replaced with '<blob>' in the log.
	    '''
	    if all([ x in printable for x in c]): return c
	    else: return '<blob>'

	no_issuer = []
	# Load each credential using the loaders.  Note that credentials
	# imported before the identity certificate that signed them are
	# unprocessable.  We put them aside and retry after the rest of the
	# batch has been imported.
	for c in creds:
	    if load_id(c) != ABAC.ABAC_CERT_SUCCESS:
		a_rv = load_attribute(c) 
		if a_rv == ABAC.ABAC_CERT_MISSING_ISSUER:
		    no_issuer.append(c)
		    if log: log.debug("Deferred %s" % show_cred(c))
		elif a_rv != ABAC.ABAC_CERT_SUCCESS:
		    if log: log.warn("Cannot import credential " + \
			    "into ABAC context: %s" % show_cred(c))
		else:
		    if log: log.debug("Imported %s" % show_cred(c))
	    else:
		if log: log.debug("Imported %s" % show_cred(c))
	# Re-process the deferred credentials
	for c in no_issuer:
	    if load_attribute(c) != ABAC.ABAC_CERT_SUCCESS:
		if log: log.warn("Cannot import credential into " + \
			"ABAC context: %s" % show_cred(c))
	    else:
		if log: log.debug("Imported %s" % show_cred(c))

    @staticmethod
    def log_context(ctxt, log):
	'''
	Debugging.  Print the ABAC context to the given log.
	'''
	if log:
	    log.debug("Context dump")
	    for c in ctxt.credentials():
		log.debug("%s <- %s" % (c.head().string(), c.tail().string()))

    def verify_from_strings(self, gid, credentials, target_urn, privileges):
	'''
	Verify that the given gid can prove it has credentials for the given
	urn.  Eventaully those credentials will be expressed in RT1
	parameterized credentials, but as the current ABAC only supports RT0 we
	use a text transform to form RT0 credentials.  Specifically, privilege
	p on urn u is a credential p_u signed by this AM.  The ABAC policy may
	provide derivation paths to those credentials.

	The routine returns a list of proven privileges with a proof for each.
	That list contains abac.proof objects.
	'''

	def binary_to_data(x):
	    '''
	    Helper to convert Binary objects into their string format.
	    '''
	    if isinstance(x, xmlrpclib.Binary): return x.data
	    else: return x

        # Note that here we treat a list of credentials as being options
        # Alternatively could accumulate privileges for example
        # The semantics of the list of credentials is under specified.

        self.logger.debug('Converting to ABAC creds')
	# Clone the context
	ctxt = ABAC.Context(self.context)
	# Load the new credentials
	self.load_creds(ctxt.load_id_chunk, ctxt.load_attribute_chunk, 
		map(binary_to_data,credentials), self.logger)
	# Debugging.
	# self.log_context(ctxt, self.logger)
	principal = GID_string_to_ABAC_ID_string(gid)
	rv = [ ]
	failed = [ ]

	# Weirdly, some operations do not ask for privileges - e.g. the
	# ListResources operation.  (Actually in the GENIAPI Reference that
	# asks for neither URN nor privs, which is nonsense to ABAC, really).
	# This creates a faux RT1 credential for privs on URN (as above) or
	# just for an unadorned URN if no privs are specified.
	if privileges:
	    for pr in privileges:
		if target_urn:
		    attr = "%s.%s_%s" % \
			    (self.me, clean_attr(pr), clean_attr(target_urn))
		else:
		    attr = "%s.%s" % (self.me, clean_attr(pr))
		self.logger.debug("Checking %s for %s" % (principal,attr))
		# Here's the proving
		ok, certs = ctxt.query(attr, principal)
		self.logger.debug("result %s" % ok)
		p = proof(self.me, principal, attr, certs)

		if ok: rv.append(p)
		else: failed.append(p.to_xml())
	else:
	    # No privileges requested
	    if target_urn:
		attr = "%s.%s" % (self.me, clean_attr(target_urn))
	    else:
		# No target either!?
		fault_code = 'Badly configured policy'
		fault_string = 'No target or privileges requeted'
		raise xmlrpclib.Fault(fault_code, fault_string)

	    self.logger.debug("Checking %s for %s" % (principal,attr))
	    # Proving for naked URN
	    ok, certs = ctxt.query(principal, attr)
	    self.logger.debug("result %s" % ok)
	    p = proof(me, principal, attribute, certs)

	    if ok: rv.append(p)
	    else: failed.append(p.to_xml())

	if rv: 
	    # one or more proof attempts succeeded.  Save them for later
	    # queries and return them to the caller.
	    self.proofs[principal] = rv
	    return rv
	else:
	    # Nothing worked.  Clear any saved successes (just in case) and
	    # raise an exception that encodes the partial proofs.
	    if principal in self.proofs:
		del self.proofs[principal]
            fault_code = 'Insufficient privileges'
            fault_string = 'Failed: %s' %  (join(failed, ''))
	    raise xmlrpclib.Fault(fault_code, fault_string)

    def get_proof(self, gid_str):
	''' Retrieve a cached proof.  '''
	if gid_str in self.proofs:
	    rv = self.proofs[gid_str]
	    del self.proofs[gid_str]
	else:
	    rv = []
	return rv

    def import_creds_binary(self, creds):
	'''
	Import a set of ABAC credentials encoded as binaries into this
	validator.
	'''
	self.load_creds(self.context.load_id_chunk, 
		self.context.load_attribute_chunk, 
		creds, self.logger)

    def import_creds_files(self, creds):
	'''
	Import a set of ABAC credentials encoded as filenames into this
	validator.
	'''
	self.load_creds(self.context.load_id_file, 
		self.context.load_attribute_file, 
		creds, self.logger)

# Utilities

def make_abac_ID(cert, key, dir=None, filename=None):
    '''
    The creddy code expects self-signed identities.  This derives a self-signed
    identity from the given certificate key pair.  The identity has the same
    (ignored) subject and duration (not ignored) as the original cert and binds
    to the same key, but is appropriate for creddy.
    If dir is given, the new ID will be created there; if not a temporary
    directory will be created.  If filename is given, it will have that name. 
    Caller is responsible for cleanup of files and any created tempdir.
    '''
    # Make sure openssl is installed
    if openssl_path is None:
	raise ValueError("Couldn't find openssl.  Is it installed?")

    if dir: 
	d = dir
	delete_dir_on_fail = False
    else:
	d = mkdtemp()
	delete_dir_on_fail = True

    fb = filename or 'abac_ID.pem'
    fn = '%s/%s' % (d, fb)

    c = X509.load_cert(cert)
    subj = c.get_subject().as_text()

    if subj.startswith('/'): i = 1
    else: i = 0
    subj = '/' + re.sub('/', '\/', subj[i:])

    dt = c.get_not_after().get_datetime()
    now = datetime.datetime.now(dt.tzinfo)
    days = dt - now

    try:
	cmd = [openssl_path, 'req', '-new', '-nodes', '-subj', subj, '-x509', 
		'-days',  '%s' % min(1, days.days), '-key', key, '-out', fn] 
	rv = subprocess.call(cmd)
	if rv != 0:
	    return None
    except Exception, e:
	if delete_dir_on_fail:
	    if os.access(fn, os.F_OK): os.unlink(fn)
	    os.rmdir(d)
	raise e

    return fn

def call_creddy_attribute(issuer_keyfile, issuer_certfile, role, life_days,
	subject_cert=None, subject_id=None, subject_role=None, 
	subject_cert_is_issuer=False):
    '''
    Call creddy to make an attribute certificate that expressed the given ABAC
    rule or attribute.  The parameters are very similar to the arguments to
    creddy (see http://abac.deterlab.net/wiki/Creddy).  This routine is called
    internally to deam with many vaguaries of ABAC credential generation in one
    place.  Note that only one of subject_cert_is_issuer, subject_cert, or
    subject_id will be honored, and they are considered in that order.  If
    subject_cert_is_issuer is false, the subject_cert is considered; if that is
    empty, the subject_id is considered.

    The contents of the generated DER-encoded X.509 attribute certificate are
    returned.
    '''
    # Make a tmeporary directory for creddy to store data.  May raise an
    # exception
    tmpdir = mkdtemp()
    cred_file = os.path.join(tmpdir, 'cert.der')

    attr_certfile = issuer_certfile
    try:
	# Confirm that the certificate can be used by creddy, and make a
	# temporary if not.
	creddy_cmd = [ creddy_path, '--verify', '--cert', issuer_certfile ]
	err_f = sys.stderr
	if os.access('/dev/null', os.W_OK):
	    err_f = open('/dev/null', 'w')
	rv = subprocess.call(creddy_cmd, stdout=err_f, stderr=err_f)
	if err_f != sys.stderr: err_f.close()
	if rv != 0:
	    attr_certfile = make_abac_ID(issuer_certfile, issuer_keyfile, 
		    dir=tmpdir)
	    if not attr_certfile:
		raise ValueError('Cannot make an ABAC ID from %s' % \
			issuer_certfile)


	# Compose the creddy command
	creddy_cmd = [creddy_path, '--attribute', '--issuer', attr_certfile, 
		'--key', issuer_keyfile, 
		'--role', role, '--out', cred_file]

	if subject_cert_is_issuer:
	    creddy_cmd.extend([ '--subject-cert', attr_certfile])
	elif subject_cert:
	    creddy_cmd.extend([ '--subject-cert', subject_cert])
	elif subject_id:
	    creddy_cmd.extend(['--subject-id', subject_id])

	if subject_role:
	    creddy_cmd.extend(['--subject-role', subject_role])

	# Call out to creddy (soon to be replaced with a library) and clean up
	# the temporaries.
	rv = subprocess.call(creddy_cmd)
	if rv != 0:
	    return None

	f = open(cred_file, 'r')
	cred = f.read()
	f.close()
	return cred

    finally:
	# All paths out of the try block come through here, to clean up.
	if os.access(cred_file, os.F_OK):
	    os.unlink(cred_file)
	if issuer_certfile != attr_certfile and os.access(attr_certfile, 
		os.F_OK):
	    os.unlink(attr_certfile)
	os.rmdir(tmpdir)

def create_delegation(trust_delegation, target_attr, issuer_keyfile, 
	issuer_certfile, life_secs=60*60*24*14):
    '''
    Create and Return a DER representation of a credential that delegates the
    target_attr(in the issuer's namespace) to the trust_delegation attribute.
    This builds an indirect delegation.  Bad arguments cause a value error to
    be raised, and an internal tempfile creation may cause various
    EnvironmentErrors to be raised.
    '''
    # Creddy thinks in days.
    life_days = max(1, life_secs / (60*60*24)) 
    return call_creddy_attribute(issuer_keyfile, issuer_certfile, target_attr, 
	    life_days, subject_cert_is_issuer=True, 
	    subject_role=trust_delegation)


def create_credential(caller_gid, object_gid, life_secs, typename, 
	issuer_keyfile, issuer_certfile, trusted_roots):
    '''
    Create and Return a DER representation of a credential that grants the
    permission in typename with respect to the URN encoded in the object_gid to
    the caller_gid, from the issuer.  Trusted roots is ignored, and this
    routine accepts any typename.  Bad arguments cause a value error to be
    raised, and an internal tempfile creation may cause various
    EnvironmentErrors to be raised.
    '''

    # Make sure cerddy is installed
    if creddy_path is None:
	raise ValueError("Couldn't find creddy.  Is it installed?")

    # Make sure everything we need came in
    if not all((caller_gid, object_gid, issuer_keyfile, typename, 
	    issuer_certfile)):
	raise ValueError("Required parameter is missing.  (required: " + \
		"caller_gid, object_gid, issuer_keyfile, typename, " +\
		"issuer_certfile")

    # Make sure some duration came in as well
    if not isinstance(life_secs, int) or life_secs < 0:
	raise ValueError("Bad duration: %s" % life_secs)

    # Use creddy to create the credential

    # Creddy thinks in days.
    life_days = max(1, life_secs / (60*60*24)) 

    # Convert to ABAC identites and convert the target/tyep to a single string.
    # As above, when we have RT1 this will be cleaner.
    principal_abac = GID_to_ABAC_ID_string(caller_gid)
    target_abac = clean_attr(object_gid.get_urn())
    priv_abac = clean_attr(typename)
    abac_attr = '%s_%s' % (priv_abac, target_abac)

    return call_creddy_attribute(issuer_keyfile, issuer_certfile, abac_attr, 
	    life_days, subject_id=principal_abac)

def clean_attr(s):
    '''
    ABAC attributes currently can only contain letters numbers and underscores.
    This replaces other characters with underscores and returns the cleaned
    string.
    '''
    return re.sub('[^A-Za-z0-9_]+', '_', s)

def m2key_to_ABAC_ID_string(pk):
    '''
    Turn a public key (in the format of M2Crypto) into an ABAC principal
    string.  In particular, calculate the sha1 hash of the bits of the key.
    '''
    # Get the key bits.  This is a bit roundabout.  We get the key
    # as a DER, then decode that into ASN1 which gives us a literal bit
    # string.
    asn1 = decoder.decode(pk.as_der())
    bits = asn1[0].getComponentByPosition(1)
    # Now we crunch that all back into a string of bytes.  Walk through
    # the bits 8 at a time, and convert them to integers, then back to
    # characters.
    bytes = ''
    for i in range(0, len(bits), 8):
	int_str = join(["%d" % b for b in bits[i:i+8]], '')
	byte = chr(int(int_str, 2))
	bytes += byte

    # Now bytes is a string with the bits from the pubkey in it.
    # Compute the hash.
    dig = MessageDigest('sha1')
    dig.update(bytes)
    hash = dig.final()

    # Now, we convert the bytes in the hash into a string of 2-dugit
    # hex strings and return it
    return join(['%02x' % ord(h) for h in hash], '')

def GID_to_ABAC_ID_string(g):
    """
    Convert a GID into an ABAC principal ID.  The ABAC principal ID is the
    sha1 hash of the public key bits, which we return as a string.  Getting
    to those is a bit of a roundabout proposition.  This is longer than it
    has to be but should be clear.
    """

    if g: return m2key_to_ABAC_ID_string(g.get_pubkey().m2key)
    else: return None

def GID_string_to_ABAC_ID_string(g_string):
    """
    Convert the string representation of a SID into an ABAC principal ID.
    """

    return GID_to_ABAC_ID_string(gid.GID(string=g_string))

