#!/usr/local/bin/python
#----------------------------------------------------------------------
# Copyright (c) 2010-2011 USC/ISI
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and/or hardware specification (the "Work") to deal in the
# Work without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Work, and to permit persons to whom the Work is furnished to do so,
# subject to the following conditions: The above copyright notice and this
# permission notice shall be included in all copies or substantial portions of
# the Work.
# 
# 
# THE WORK IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE WORK OR THE USE OR OTHER DEALINGS
# IN THE WORK.
#----------------------------------------------------------------------
import os

import logging
import traceback
import xmlrpclib

from tempfile import mkdtemp
from string import join

from M2Crypto import SSL
from M2Crypto.SSL.SSLServer import SSLServer
from BaseHTTPServer import BaseHTTPRequestHandler

from abac.cred_util import CredentialVerifier

# ABAC IDs are not tied to a hierarchy of trust, they are just key hashes.
# This is an XMLRPC server that accepts self-signed requests, used to call the
# ABAC clients.  It will also work fine with a trusted hierarchy as well, if
# need be.

class ssl_context(SSL.Context):
    """
    Simple wrapper around an M2Crypto.SSL.Context that allows validation of
    self signed certificates.  This is also used by fedd
    (http://fedd.deterlab/net) for the same purpose.
    """
    def __init__(self, my_cert, my_key, trusted_certs=None, password=None):
	"""
	@param my_cert: PEM file with my certificate in it
	@param trusted_certs: PEM file with trusted certs in it (optional)
	"""

	def cleanup(tmpdir, certfile):
	    if tmpdir:
		if certfile and os.access(certfile, os.F_OK):
		    os.unlink(certfile)
		os.rmdir(tmpdir)
		delete_certfile = None

	SSL.Context.__init__(self)


	# Exceptions propagate out of the try, but the cleanup gets done.
	delete_certfile = False
	try:
	    # If a key file is given we need to make a combo file for M2Crypto.
	    # This copies key and cert into a tempfile that is removed after
	    # loading
	    if my_key:
		tmpdir = mkdtemp()
		certfile = "%s/cert.pem" % tmpdir
		# This is a very compressed way to read a file into a string.
		key_str = join([ line for line in open(my_key, 'r')],'')
		self.cert_str = join([ line for line in open(my_cert, 'r')],'')
		# Write the combo file
		out = open(certfile, 'w')
		print >>out, key_str
		print >>out, self.cert_str
		out.close()
		delete_certfile = True
	    else:
		# A combo was passed in (we hope) do not delete anything.
		self.cert_str = join([ line for line in open(my_cert, 'r')],'')
		certfile = my_cert
		tmpdir = None


	    # load_cert takes a callback to get a password, not a password, so
	    # if the caller provided a password, this creates a nonce callback
	    # using a lambda form.
	    if password != None and not callable(password):
		# This is cute.  password = lambda *args: password produces a
		# function object that returns itself rather than one that
		# returns the object itself.  This is because password is an
		# object reference and after the assignment it's a lambda.  So
		# we assign to a temp.
		pwd = str(password)
		password =lambda *args: pwd

	    if password != None:
		self.load_cert(certfile, callback=password)
	    else:
		self.load_cert(certfile)
	finally:
	    if delete_certfile: cleanup(tmpdir, certfile)
	
	# If no trusted certificates are specified initialize them, if not
	# allow unknown CAs.
	if trusted_certs: 
	    self.load_verify_locations(trusted_certs)
	    self.set_verify(SSL.verify_peer, 10)
	else:
	    callb = SSL.cb.ssl_verify_callback_allow_unknown_ca
	    self.set_allow_unknown_ca(True)
	    self.set_verify(SSL.verify_peer, 10, callback=callb)

class Server(SSLServer):
    """
    Interface the ABAC Controlled services to XMLRPC
    """
    def __init__(self, ME, delegate, keyfile, certfile, ca_certs, base_name,
	    debug=False):
	"""
	Create an SSL server that handles the transport in handler using the
	credentials in ssl_ctx, and interfacing to the implementation of fedd
	services in fedd.  ME is the host port pair on which to bind.
	"""
	ssl_ctx = ssl_context(certfile, keyfile, ca_certs)
	SSLServer.__init__(self, ME, xmlrpc_handler, ssl_ctx)
	self.log = logging.getLogger("gcf-am.reference")
	self.delegate = delegate
	self.me_pem = ssl_ctx.cert_str
	self.me_key = keyfile
	self.me_certfile = certfile
	delegate.set_server(self)
	if debug: self.handle_error = self.handle_error_debug
	else: self.handle_error = self.handle_error_standard

    def handle_error_debug(self, request=None, client_address=None):
	print '-'*40
	traceback.print_exc()
	print '-'*40


    def handle_error_standard(self, request=None, address=None):
	"""
	The default SSLServer prints a stack trace here.  This is a little
	friendlier.
	"""
	if request or address:
	    self.log.warn("Error on incoming connection: %s %s" % \
		    (request or "", address or ""))
	else:
	    self.log.warn("Error on incoming connection (Likely SSL error)")

class xmlrpc_handler(BaseHTTPRequestHandler):
    """
    Handle XMLRPC requests.

    Much of this is boilerplate from 
    http://www.xml.com/pub/a/ws/2004/01/20/salz.html
    """
    server_version = "GCF/1.0 ABAC/1.2 " + BaseHTTPRequestHandler.server_version

    def send_xml(self, text, code=200):
	"""Send an XML document as reply"""
	self.send_response(code)
	self.send_header('Content-type', 'text/xml; charset="utf-8"')
	self.send_header('Content-Length', str(len(text)))
	self.end_headers()
	self.wfile.write(text)
	self.wfile.flush()
	# Make sure to close the socket when we're done
	self.request.socket.close()

    def do_POST(self):
	"""Treat an HTTP POST request as an XMLRPC service call"""
	# NB: XMLRPC faults are not HTTP errors, so the code is always 200,
	# unless an HTTP error occurs, which we don't handle.

	cl = int(self.headers['content-length'])
	data = self.rfile.read(cl)

	try:
	    params, method = xmlrpclib.loads(data)
	except xmlrpclib.ResponseError:
	    data = xmlrpclib.dumps(xmlrpclib.Fault("Client", 
		"Malformed request"), methodresponse=True)

	# tuck the caller into pem_cert, which is where the AM and CH look for
	# it.  That will have to change to be thread safe.
	self.server.pem_cert = self.request.get_peer_cert().as_pem()

	if method != None:
	    try:
		method_call = getattr(self.server.delegate, method, None)
		if method_call:
		    resp = method_call(*params)
		    data = xmlrpclib.dumps((resp,), encoding='UTF-8', 
			    methodresponse=True)
		else:
		    raise xmlrpclib.Fault(100, "Unknown method: %s" % method)
	    except xmlrpclib.Fault, f:
		data = xmlrpclib.dumps(f, methodresponse=True)
		resp = None

	self.send_xml(data)

    def log_request(self, code=0, size=0):
	""" Log request outcome.  """
	self.server.log.info("Successful XMLRPC request code %d" % code)
